from bacon import baconNet
import numpy as np

import sys
sys.path.append("..")


class lsp3(baconNet):
    def __init__(self):
        super().__init__(3, constantTerm=False)

    def get_contribution(self, a, b):
        return super().get_contribution(a, b)

    def explain(self, singleVariable=False):
        if singleVariable:
            return "A"
        m, c = self.get_contribution(1, 1)
        delta = 0.01
        terms = []
        if abs(m[0][0]-1) < delta:
            return "min(A, B)"
        if abs(m[1][0]-1) < delta:
            return "max(A, B)"
        if abs(m[0][0]-0.5) < delta and abs(m[1][0]-0.5) < delta:
            return "(A + B) / 2"
        if abs(m[2][0]-1) < delta:
            return "A * B"
        return "can't explain"

    def expand(self, a, b, y):
        X = np.column_stack((np.minimum(a, b),
                             np.maximum(a, b),
                             np.product(np.array([a, b]), axis=0)))
        return np.column_stack((a, b)), X, y



from abc import abstractmethod


from abc import ABC
import tensorflow as tf
from tensorflow import keras
from keras.utils.generic_utils import get_custom_objects
from keras.layers import Activation
import numpy as np

# surpress warnings on running GPU Tensorflow on CPU
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


class baconNet(ABC):
    def __custom_activation(self, x):
        return x

    def __init__(self, size, constantTerm=True):
        get_custom_objects().update(
            {'custom_activation': Activation(self.__custom_activation)})
        self.__model = tf.keras.models.Sequential()
        self.__model.add(tf.keras.Input(shape=(size,)))
        self.__model.add(tf.keras.layers.Dense(
            size, activation=self.__custom_activation))
        self.__model.add(tf.keras.layers.Dense(
            1, activation=self.__custom_activation))
        self.__model.compile(
            optimizer='adam', loss='mean_squared_error', metrics=['mae'])
        self.__size = size
        self.__constantTerm = constantTerm

    @abstractmethod
    def explain(self, singleVariable=False):
        pass

    @abstractmethod
    def get_contribution(self, a, b):
        m = np.matmul(
            self.__model.layers[0].weights[0], self.__model.layers[1].weights[0])
        if self.__constantTerm:
            # assuming the last input is a constant column
            bias0 = self.__model.layers[0].bias.numpy()
            weight0 = self.__model.layers[0].weights[0][self.__size-1]
            for i in range(len(bias0)):
                bias0[i] = bias0[i] + weight0[i]
            weight1 = self.__model.layers[1].weights[0]
            for i in range(len(bias0)):
                bias0[i] = bias0[i] * weight1[i][0]
            c = bias0.sum() + self.__model.layers[1].bias.numpy()[0]
        else:
            c = 0
        return m, c

    @abstractmethod
    def expand(self, a, b, y):
        pass

    def predict(self, a, b=0):
        simple = False
        if isinstance(a, list):
            if isinstance(b, list):
                x, X, y = self.expand(a, b, np.zeros(len(a)))
            else:
                x, X, y = self.expand(a, np.zeros(len(a)), np.zeros(len(a)))
        else:
            simple = True
            x, X, y = self.expand([a], [b], [0])
        if simple:
            return self.__model.predict(X)[0][0]
        else:
            return self.__model.predict(X).flatten()

    def fit(self, a, b, y):
        callback = tf.keras.callbacks.EarlyStopping(monitor='mae', patience=20)
        x, X, y = self.expand(a, b, y)
        history = self.__model.fit(
            X, y, epochs=600, batch_size=10, verbose=0, callbacks=[callback])
        return history


class dataCreator:
    def create(size, scale, aggregate, singleVariable=False):
        a = np.random.rand(size) * scale
        if not singleVariable:
            b = np.random.rand(size) * scale
        else:
            b = np.zeros(size)
        y = aggregate(a, b)
        return a, b, y



from bacon import baconNet
import numpy as np

import sys
sys.path.append("..")


class poly2(baconNet):
    def __init__(self):
        super().__init__(6)

    def get_contribution(self, a, b):
        return super().get_contribution(a, b)

    def explain(self, singleVariable=False):
        a = 1
        b = 1
        if singleVariable:
            b = 0
        m, c = self.get_contribution(a, b)
        delta = 0.01
        terms = []
        if abs(m[2][0]) > delta and abs(a) > delta:
            terms.append(str(round(m[2][0], 4)) + "x^2")
        if abs(m[4][0]) > delta and abs(b*a) > delta:
            terms.append(str(round(m[4][0], 4)) + "xy")
        if abs(m[3][0]) > delta and abs(b) > delta:
            terms.append(str(round(m[3][0], 4)) + "y^2")
        if abs(m[0][0]) > delta and abs(a) > delta:
            terms.append(str(round(m[0][0], 4)) + "x")
        if abs(m[1][0]) > delta and abs(b) > delta:
            terms.append(str(round(m[1][0], 4)) + "y")
        if abs(m[5][0]) > delta and abs(c) > delta:
            terms.append(str(round(c, 4)))
        return "z = " + " + ".join(terms)

    def expand(self, a, b, y):
        X = np.column_stack((a, b,
                             np.power(a, 2),
                             np.power(b, 2),
                             np.product(np.array([a, b]), axis=0),
                             np.ones(len(a))))
        return np.column_stack((a, b)), X, y
